use std::{
    collections::{HashMap, VecDeque},
    net::SocketAddr,
    sync::Arc,
};

use tokio::{select, sync::Semaphore};
use tokio_util::sync::CancellationToken;

use crate::{
    atem_lib::atem_socket::{AtemEvent, AtemSocketCommand, AtemSocketMessage, TrackingId},
    commands::{
        command_base::{BasicWritableCommand, DeserializedCommand},
        device_profile::DESERIALIZE_VERSION_RAW_NAME,
        init_complete::DESERIALIZE_INIT_COMPLETE_RAW_NAME,
        time::DESERIALIZE_TIME_RAW_NAME,
    },
    state::AtemState,
};

#[derive(Default)]
pub enum AtemConnectionStatus {
    #[default]
    Closed,
    Connecting,
    Connected,
}

pub struct Atem {
    waiting_semaphores: tokio::sync::RwLock<HashMap<TrackingId, Arc<Semaphore>>>,
    socket_message_tx: tokio::sync::mpsc::Sender<AtemSocketMessage>,
}

impl Atem {
    pub fn new(socket_message_tx: tokio::sync::mpsc::Sender<AtemSocketMessage>) -> Self {
        Self {
            waiting_semaphores: tokio::sync::RwLock::new(HashMap::new()),
            socket_message_tx,
        }
    }

    pub async fn connect(&self, address: SocketAddr) -> bool {
        let (callback_tx, callback_rx) = tokio::sync::oneshot::channel();
        self.socket_message_tx
            .send(AtemSocketMessage::Connect {
                address,
                result_callback: callback_tx,
            })
            .await
            .unwrap();

        callback_rx.await.unwrap()
    }

    pub async fn run(
        &self,
        mut atem_event_rx: tokio::sync::mpsc::UnboundedReceiver<AtemEvent>,
        cancel: CancellationToken,
    ) {
        let mut status = AtemConnectionStatus::default();
        let mut state = AtemState::default();

        while !cancel.is_cancelled() {
            select! {
                _ = cancel.cancelled() => {},
                message = atem_event_rx.recv() => match message {
                    Some(event) => match event {
                        AtemEvent::Connected => {
                            log::info!("Atem connected");
                        }
                        AtemEvent::Disconnected => todo!("Disconnected"),
                        AtemEvent::ReceivedCommands(commands) => {
                            self.mutate_state(&mut state, &mut status, commands).await
                        }
                        AtemEvent::AckedCommand(tracking_id) => {
                            log::debug!("Received tracking Id {tracking_id}");
                            if let Some(semaphore) =
                                self.waiting_semaphores.read().await.get(&tracking_id)
                            {
                                semaphore.add_permits(1);
                            } else {
                                log::warn!("Received tracking Id {tracking_id} with no-one waiting for it to be resolved.")
                            }
                        }
                    },
                    None => {
                        log::info!("ATEM event channel has closed, exiting event loop.");
                        cancel.cancel();
                    }
                }
            }
        }
    }

    pub async fn send_commands(&self, commands: Vec<Box<dyn BasicWritableCommand>>) {
        let (callback_tx, callback_rx) = tokio::sync::oneshot::channel();
        self.socket_message_tx
            .send(AtemSocketMessage::SendCommands {
                commands: commands
                    .iter()
                    .map(|command| {
                        AtemSocketCommand::new(command, &crate::enums::ProtocolVersion::Unknown)
                    })
                    .collect(),
                tracking_ids_callback: callback_tx,
            })
            .await
            .unwrap();
        let callback = callback_rx.await.unwrap();

        let semaphore = Arc::new(Semaphore::new(0));

        for tracking_id in callback.tracking_ids.iter() {
            self.waiting_semaphores
                .write()
                .await
                .insert(tracking_id.clone(), semaphore.clone());
        }

        callback.barrier.wait().await;

        // If this fails then the semaphore has been closed which is a darn shame but at that point
        // the best we can do it continue on our merry way in life and remain oblivious to
        // the fire raging in other realms.
        semaphore
            .acquire_many(callback.tracking_ids.len() as u32)
            .await
            .ok();

        for tracking_id in callback.tracking_ids.iter() {
            self.waiting_semaphores.write().await.remove(tracking_id);
        }
    }

    async fn mutate_state(
        &self,
        state: &mut AtemState,
        status: &mut AtemConnectionStatus,
        commands: VecDeque<Arc<dyn DeserializedCommand>>,
    ) {
        let new_state = state.clone();
        for command in commands {
            match command.raw_name() {
                DESERIALIZE_VERSION_RAW_NAME => {
                    log::debug!("Received version response");
                    *state = AtemState::default();
                    *status = AtemConnectionStatus::Connecting
                }
                DESERIALIZE_INIT_COMPLETE_RAW_NAME => {
                    log::debug!("Received init complete from ATEM");
                    *status = AtemConnectionStatus::Connected
                }
                DESERIALIZE_TIME_RAW_NAME => {
                    todo!("Time command")
                }
                _ => {
                    log::debug!("Applying {} to state", command.raw_name());
                    command.apply_to_state(state);
                }
            }
        }

        if new_state != *state {
            *state = new_state;
            todo!("Emit change");
        }
    }
}
